/*
	Crossing
	Copyright (C) 2007 Michal Vaner (michal.vaner@kdemail.net)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	he Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef BEHAV_H
#define BEHAV_H

#include <QList>
#include <QVector>
#include <QDomElement>

class Vehicle;
class Point;
class Line;
class VehicType;
class SemaphoreGroup;
class Indexer;
class SemaphoreBehav;

struct VehicState {
	int speed;
	int priority;
	Line *cont;
	QList<const SemaphoreBehav *> semaphores;
	VehicState();
};

class Behavior {
	public:
		virtual ~Behavior();
		virtual VehicState processVehicle(const VehicState &state, const Vehicle &vehicle) const = 0;
};

class Condition : public Behavior {
	protected:
		QList<Condition *> conditions;//Sub-conditions
		QList<Behavior *> behaviors;//All other condition modifiers
	private:
		class Cleaner {
			private:
				Condition *mCondition;
			public:
				Cleaner(Condition *condition) : mCondition(condition) {}
				~Cleaner();
		} cleaner;
	public:
		Condition(const QDomElement &node, Point *point);
		virtual VehicState processVehicle(const VehicState &state, const Vehicle &vehicle) const;
	protected:
		virtual bool match(const Vehicle &vehicle) const = 0;
};

class RootCond : public Condition {
	protected:
		virtual bool match(const Vehicle &vehicle) const;
	public:
		RootCond(const QDomElement &node, Point *point);
};

class VehicCond : public Condition {
	private:
		const VehicType *type;
	public:
		VehicCond(const QDomElement &node, Point *point);
		virtual bool match(const Vehicle &vehicle) const;
};

class FromCond : public Condition {
	private:
		Line *line;
	public:
		FromCond(const QDomElement &element, Point *point);
		virtual bool match(const Vehicle &vehicle) const;
};

class SpeedBehav : public Behavior {
	private:
		int speed;
	public:
		SpeedBehav(const QDomElement &node);
		SpeedBehav();
		virtual VehicState processVehicle(const VehicState &state, const Vehicle &vehicle) const;
};

class ContBehav : public Behavior {
	private:
		int max;
		struct Cont {
			int start, len;
			Line *line;
		};
		QVector<Cont> conts;
	public:
		ContBehav(const QDomElement &node, Point *point);
		ContBehav(Point *point);//Each one of them with the same probability (as the default)
		void addMore(const QDomElement &node, Point *point);
		VehicState processVehicle(const VehicState &state, const Vehicle &vehicle) const;
};

class SemaphoreBehav : public Behavior {
	private:
		SemaphoreGroup *group;
		struct Interval {
			int start, end;
			bool operator<(const Interval &other) const { return ((start < other.start) || ((start == other.start) && (end < other.end))); }
		};
		QVector<Interval> intervals;
	public:
		SemaphoreBehav(const QDomElement &node, Indexer *indexer);
		VehicState processVehicle(const VehicState &state, const Vehicle &vehicle) const;
		bool isOpen() const;
};

class PriorityBehav : public Behavior {
	private:
		int priority;
	public:
		PriorityBehav(const QDomElement &node);
		PriorityBehav();
		VehicState processVehicle(const VehicState &state, const Vehicle &) const;
};

#endif//BEHAV_H
