/*
	Crossing
	Copyright (C) 2007 Michal Vaner (michal.vaner@kdemail.net)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	he Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "crossget.h"
#include "indexer.h"
#include "point.h"
#include "line.h"
#include "vehicle.h"
#include "mapobj.h"

#include <QPainter>
#include <QPaintEvent>
#include <QBrush>

Crossget::Crossget(QWidget *parent) :
	QWidget(parent),
	indexer(NULL),
	mDrawStruct(false)
{
}

Crossget::~Crossget() {
	if(indexer)
		indexer->deleteLater();
}

void Crossget::setBackground(const QPixmap &_pixmap) {
	if(indexer) {
		indexer->deleteLater();
		indexer = NULL;
	}
	pixmap = _pixmap;
	resize(pixmap.width(), pixmap.height());
	update();
}

void Crossget::paintEvent(QPaintEvent *event) {
	QPainter painter(this);
	painter.setClipRegion(event->region());
	painter.drawPixmap(0, 0, pixmap);
	if(!indexer)
		return;//Nothing to show
	foreach(const Indexer::Level *level, indexer->levels()) {
		QList<const Indexable *> objects = level->objects();
		foreach(const Indexable *obj, objects) {
			const MapObject *o = dynamic_cast<const MapObject *>(obj);
			painter.save();
			painter.translate(o->x(), o->y());
			painter.rotate(o->rot());
			painter.drawPixmap(-o->pixmap().width() / 2, -o->pixmap().height() / 2, o->pixmap());
			painter.restore();
		}
		if(mDrawStruct) {
			painter.setBrush(QBrush(Qt::red));
			QPen pen(Qt::red);
			pen.setWidth(3);
			painter.setPen(pen);
			foreach(const Indexable *i, level->structurals()) {
				const Point *p = dynamic_cast<const Point *>(i);
				if(p)
					painter.drawEllipse(p->x() - 5, p->y() - 5, 10, 10);
				else {
					const Line *l = dynamic_cast<const Line *>(i);
					if(l)
						painter.drawLine(l->src()->x(), l->src()->y(), l->dest()->x(), l->dest()->y());
				}
			}
		}
		QList<const Indexable *> vehicles = level->vehicles();
		foreach(const Indexable *vehicI, vehicles) {
			const Vehicle *v = dynamic_cast<const Vehicle *>(vehicI);
			painter.save();
			const QPixmap pm = v->img();
			painter.translate(v->x(), v->y());
			painter.rotate(v->rotation());
			painter.scale(2.0/3.0, 2.0/3.0);//Scale down to decrease pixel effects
			painter.drawPixmap(-pm.width() / 2, -pm.height() / 2, pm);
			painter.restore();
		}
	}
}

void Crossget::insertIndexer(Indexer *indexer) {
	if(this->indexer)
		this->indexer->deleteLater();
	this->indexer = indexer;
	connect(indexer, SIGNAL(timeTick(int)), SLOT(update()), Qt::QueuedConnection);
}

void Crossget::setDrawStruct(bool value) {
	if(value != mDrawStruct) {
		mDrawStruct = value;
		update();
	}
}
