/*
	Crossing
	Copyright (C) 2007 Michal Vaner (michal.vaner@kdemail.net)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	he Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "indexer.h"
#include "load.h"
#include "point.h"
#include "line.h"
#include "semaphore_group.h"
#include "vehicle.h"

#include <QtAlgorithms>
#include <QTimer>
#include <assert.h>

Indexable::Indexable(Type type, int index) :
	mIndex(index),
	mType(type),
	mIndexer(NULL)
{
}

Indexable::~Indexable() {
	if(indexer())
		indexer()->removeIndexable(this);
}

int Indexable::index() const {
	return mIndex;
}

void Indexable::changeIndex(int newIndex) {
	if(newIndex != index()) {
		indexer()->removeIndexable(this);
		mIndex = newIndex;
		indexer()->insertIndexable(this);
	}
}

bool lessThan(const Indexer::Level *_1, const Indexer::Level *_2) {
	return *_1 < *_2;
}

void Indexer::insertToLevel(const Indexable *indexable, int level) {
	Level *l = mLevelIndex.value(level);
	if(!l) {
		mLevels << (l = new Level(level));
		qSort(mLevels.begin(), mLevels.end(), lessThan);
		mLevelIndex.insert(level, l);
	}
	l->mIndexables << indexable;
}

void Indexer::insertIndexable(Indexable *indexable) {
	mIndexables << indexable;
	indexable->mIndexer = this;//Let it know who is its indexer
	insertToLevel(indexable, indexable->index());
}

Indexer::~Indexer() {
	qDeleteAll(mIndexables);
	qDeleteAll(mLevelIndex.values());
}

Indexer::Indexer() :
	tickTimer(new QTimer(this)),
	mSpeed(10),
	mTraffic(500)
{
	connect(tickTimer, SIGNAL(timeout()), SLOT(getTimeDelta()));
}

QList<Point *> Indexer::points() {
	QList<Point *> result;
	Point *p;
	foreach(Indexable *i, mIndexables)
		if((p = dynamic_cast<Point *>(i)))
			result << p;
	return result;
}

void Indexer::insertSemaphoreGroup(SemaphoreGroup *semaphoreGroup) {
	mSemaphoreGroups.insert(semaphoreGroup->id(), semaphoreGroup);
	insertIndexable(semaphoreGroup);//make sure it gets deleted as well
}

void Indexer::getTimeDelta() {
	QTime act(QTime::currentTime());
	int delta = lastTime.msecsTo(act);
	if(delta > 0) {//ignore midnights
		emit preTick(delta);
		emit timeTick(delta);
		timeDiff += delta;
		while(timeDiff >= 1000 / speed()) {
			timeDiff -= 1000 / speed();
			emit tick();
		}
	}
	lastTime = act;
}

void Indexer::start() {
	lastTime = QTime::currentTime();
	timeDiff = 0;
	tickTimer->start(30);
}

void Indexer::stop() {
	tickTimer->stop();
	const QList<Indexable *> vehics = vehicles();
	foreach(Indexable *i, vehics) {
		Vehicle *v = dynamic_cast<Vehicle *>(i);
		assert(v);
		v->deleteLater();
	}
	const QList<Indexable *> strs = filter(Indexable::Structural);
	foreach(Indexable *i, strs) {
		Line *l = dynamic_cast<Line *>(i);
		if(l)
			l->clearOngoings();
	}
	emit clearPlan();
}

void Indexer::pause() {
	tickTimer->stop();
}

void Indexer::removeIndexable(Indexable *indexable) {
	mIndexables.removeAll(indexable);
	mLevelIndex.value(indexable->index())->mIndexables.removeAll(indexable);
}

void Indexer::setSpeed(int value) {
	int x[] = {7, 10, 15};
	mSpeed = x[value];
}

void Indexer::setTraffic(int value) {
	int x[] = {1000, 500, 250, 100};
	mTraffic = x[value];
}
