/*
	Crossing
	Copyright (C) 2007 Michal Vaner (michal.vaner@kdemail.net)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	he Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "point.h"
#include "load.h"
#include "globals.h"
#include "line.h"
#include "behav.h"
#include "vehicle.h"

#include <QCoreApplication>
#include <assert.h>
#include <QtAlgorithms>

Point::Point(const QDomElement &node, Indexer &indexer) :
	//Load all reasonable data
	Indexable(Indexable::Structural, getIndex(node, QCoreApplication::translate("xml_errors", "Point level error"))),
	mId(nonEmpty(node.attribute("id"), XMLLoadDataE(QCoreApplication::translate("xml_errors", "Point ID error")))),
	mNode(node),
	behavior(NULL),
	generator(NULL),
	lockDecided(false),
	lockedBy(NULL)
{
	bool ok1, ok2;
	mX = node.attribute("x").toInt(&ok1);
	mY = node.attribute("y").toInt(&ok2);
	if(!ok1 || !ok2)
		throw XMLLoadDataE(QCoreApplication::translate("xml_errors", "Point coordinate error"));
	indexer << this;
	connect(&indexer, SIGNAL(preTick(int)), SLOT(newLockRound()));//Delete the old locks every time
	connect(&indexer, SIGNAL(timeTick(int)), SLOT(clearLockReqs()));
	connect(&indexer, SIGNAL(clearPlan()), SLOT(reinitLocks()));
}

void Point::pushLine(Line *ln) {
	if(lines.contains(ln->id()))
		throw XMLLoadDataE(QCoreApplication::translate("xml_errors", "Line ID conflict"));
	lines.insert(ln->id(), ln);
}

void Point::pushBackLine(Line *ln) {
	if(inLines.contains(ln->id()))
		throw XMLLoadDataE(QCoreApplication::translate("xml_errors", "Line ID conflict"));
	assert(ln->dest() == this);//Want a back line not going here?
	inLines.insert(ln->id(), ln);
}

void Point::constructStructure() {
	behavior.reset(new RootCond(mNode, this));
	QDomNodeList lst = mNode.elementsByTagNameNS(NS, "emerge");
	for(int i = 0; i < lst.count(); i++) {
		new Generator(lst.at(i).toElement(), this);
	}
}

VehicState Point::decide(const ::Vehicle &vehicle) const {
	return behavior->processVehicle(VehicState(), vehicle);
}

void Point::clearLock(QObject *deadVehicle) {
	if(lockedBy == deadVehicle) {
		lockedBy = NULL;
	}
}

void Point::newLockRound() {
	lockDecided = false;
}

void Point::reserveLock(::Vehicle *vehicle, int priority, bool near) {
	LockRequest rq;
	rq.by = vehicle;
	rq.near = near;
	rq.priority = priority;
	lockReqs << rq;
}

bool Point::hasLock(::Vehicle *vehicle) {
	if(!lockDecided)
		decideLock();
	return lockedBy == vehicle;
}

void Point::giveUpLock(::Vehicle *vehicle) {
	if(lockedBy == vehicle)
		lockedBy = NULL;
}

bool Point::isUnlocked() {
	if(!lockDecided)
		decideLock();
	return !lockedBy;
}

void Point::decideLock() {
	if(lockReqs.count()) {
		LockRequest best = lockReqs.first();//choose the best request (it has already locked, or it is bigger priority)
		foreach(const LockRequest &req, lockReqs)
			if(better(req, best)) {//Is req better than best?
				best = req;
			}
		if(lockedBy != best.by) {
			lockedBy = best.by;//Lock it by the best one
			connect(lockedBy, SIGNAL(destroyed(QObject *)), SLOT(clearLock(QObject *)));
		}
		lockReqs.clear();//prepare for new round of requests
	} else
		lockedBy = NULL;
	lockDecided = true;
}

bool Point::better(const Point::LockRequest &_1, const Point::LockRequest &_2) const {
	if(_1.near && (_1.by == lockedBy))
		return true;
	if(_2.near && (_2.by == lockedBy))
		return false;
	return _1.priority > _2.priority;
}

void Point::clearLockReqs() {
	if(!lockDecided)
		decideLock();
}

void Point::immediateLock(::Vehicle *vehicle) {
	lockedBy = vehicle;
}

void Point::reinitLocks() {
	lockDecided = false;
	lockReqs.clear();
}
