/*
	Crossing
	Copyright (C) 2007 Michal Vaner (michal.vaner@kdemail.net)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	he Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "vehicle.h"
#include "point.h"
#include "line.h"

#include <QDir>
#include <QStringList>
#include <QFile>
#include <QTextStream>
#include <stdexcept>

static QHash<QString, const VehicType *> mVehicTypes;

const QHash<QString, const VehicType *> &vehicTypes() {
	return mVehicTypes;
}

void loadVehicTypes() {
	QDir typeDir(":/type/desc/");
	const QStringList names = typeDir.entryList();
	foreach(const QString &f, names) {
		VehicType *tp = new VehicType(f);
		mVehicTypes.insert(tp->type(), tp);
	}
}

VehicType::VehicType(const QString &fn) {
	QFile file(fn);
	if(!file.open(QIODevice::ReadOnly))
		throw std::invalid_argument("File does not exist");
	QTextStream stream(&file);
	QStringList spec = stream.readLine().split(";");
	if(spec.size() != 2)
		throw std::logic_error("Wrong vehicle desc file");
	mType = spec[0].trimmed();
	QString imgPrefix = spec[1].trimmed();
	while(!stream.atEnd()) {
		QString ln = stream.readLine().trimmed();
		if(!ln.isEmpty())
			mImageSets << new ImageSet(ln, imgPrefix);
	}
}

VehicType::ImageSet::ImageSet(const QString &ln, const QString &prefix) {
	QStringList list = ln.split(";");
	if(list.count() < 2)
		throw std::logic_error("Wrong vehicle desc file");
	bool ok;
	mImgInt = int(1000 * list.takeFirst().toFloat(&ok));
	if(!ok)
		throw std::logic_error("Invalid interval specification in vehicle desc file");
	foreach(const QString &fname, list) {
		mImgs << QPixmap(prefix + "/" + fname);
		if(mImgs.last().isNull())
			throw std::logic_error("Invalid vehicle image name %1");
	}
}

Vehicle::Vehicle(const VehicType *type, Point *point) :
	Indexable(Indexable::Vehicle, point->index()),
	imgPos(0),
	imgTime(0),
	mPoint(point),
	mLine(NULL),
	waitForVehicle(NULL),
	vtype(type),
	nextPoint(mPoint->decide(*this))//Appear in this point
{
	imageSet = vtype->sets()[random() % vtype->sets().count()];
	connect(point->indexer(), SIGNAL(timeTick(int)), SLOT(timeTick(int)));
	connect(point->indexer(), SIGNAL(preTick(int)), SLOT(canGo(int)));
	moveToPoint();//Move to this point and decide about the feature way
}

static double compDist(int time, int sp1, int sp2, double actPos, double len, int speedMulti) {
	return time * (sp1 + (sp2 - sp1) * (actPos / len)) * speedMulti / 7000;
}

void Vehicle::timeTick(int timeChange) {
	imgTime += timeChange;
	imgPos += imgTime / imageSet->imgInt();
	imgPos %= imageSet->imgs().count();
	imgTime %= imageSet->imgInt();
	if(!waiting) {
		double change = compDist(timeChange, srcSpeed, nextPoint.speed, actPos, mLine->length(), indexer()->speed());
		if((actPos + change + 35 >= mLine->length()) && !mLine->dest()->hasLock(this))
			return;//Can not go there, it is not mine
		actPos += change;
		if(actPos >= mLine->length())
			moveToPoint();
		if(actPos >= 35)
			mPoint->giveUpLock(this);//I do not want this any more
	}
}

void Vehicle::moveToPoint() {
	if(mPoint)
		mPoint->giveUpLock(this);
	if(mLine)
		mLine->leave();
	if(!nextPoint.cont)
		deleteLater();
	else {
		actPos = 0;
		if(mLine)
			mPoint = mLine->dest();
		srcSpeed = nextPoint.speed;
		mLine = nextPoint.cont;
		mLine->enter(this);
		nextPoint = mLine->dest()->decide(*this);
		changeIndex(qMax(mLine->dest()->index(), mPoint->index()));
		if(mPoint)
			mPoint->reserveLock(this, 0, true);
	}
}

double Vehicle::rotation() const {
	return mLine->angle();
}

int Vehicle::x() const {
	return mPoint->x() + int(actPos * mLine->xInc());
}

int Vehicle::y() const {
	return mPoint->y() + int(actPos * mLine->yInc());
}

void Vehicle::canGo(int timeChange) {
	double newPos = compDist(timeChange, srcSpeed, nextPoint.speed, actPos, mLine->length(), indexer()->speed()) + actPos;
	if((newPos + 25 <= mLine->length()) && (newPos + 40 >= mLine->length())) {//Check all semaphores
		foreach(const SemaphoreBehav *sem, nextPoint.semaphores)
			if(!sem->isOpen()) {
				waiting = true;
				return;
			}
	}
	waiting = waitForVehicle && (newPos + 70 > waitForVehicle->actPos);
	//TODE Semaphores
	if(actPos < 35)//block the old one
		mPoint->reserveLock(this, 0, true);
	if(!waiting && (newPos + 90 >= mLine->length()))//OK, try get the point before me
		mLine->dest()->reserveLock(this, nextPoint.priority, (actPos + 35 >= mLine->length()));
}
